/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html


  routine to compute scattering factors from a known formula and
  the tables in the International Tables for X=Ray crystallography.
  this routine should be called once and for all, since this data does 
  not change during a simulation.  

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "constants.h"
#include "memory.h"

#include "link.h"
#include "atoms.h"
#include "positions.h"
#include "chain.h"
#include "cell.h"
#include "SP_entry.h"
#include "intensity.h"
#include "spot.h"
#include "rad_pattern.h"
#include "misc.h"
#include "molecule.h"


void get_correction(struct Rad_pattern * RP,
		    const struct Positions * T,
		    const struct Cell * C, 
		    struct SP_entry ** SP)
{
  int i;
  int n_spots;

  if(RP == 0 || T == 0 || C == 0 || SP == 0) {
    push_generic_error("get_correction","(nil) args passed");
    exit_clean();
  }

  if(RP->corr == 1) {
    push_generic_error("get_correction","call this function just once");
    return;
  }

  RP->corr = 1;

  n_spots = RP->length;

  /* allocate memory for correction factors */

  if(RP->type == XRAY || RP->type == BOTH) {
    RP->xray_corr = new_vector_dp(n_spots);
    for(i = 0; i < n_spots; i++) {
      RP->xray_corr[i] = new_vector_d(T->n_unique + T->Nmol_atoms);
    }
  }

  if(RP->type == NEUTRON || RP->type == BOTH) {
    RP->neu_corr = new_vector_dp(n_spots);
    for(i = 0; i < n_spots; i++) {
      RP->neu_corr[i] = new_vector_d(T->n_unique + T->Nmol_atoms);
    }
  }

  /* loop over different spots */
  for( i = 0; i < n_spots; i++ ) {
    int o, d;
    struct Link * aa;
    double k[3], k2, sqk2, theta, pol = 0.0, npol = 0.0;
    unsigned long j;

    for(j = 0; j < 3; j ++) {
      /* this is the k vector in reciprocal space */
      k[j] = C->rec_vec[0][j] * RP->I[i]->h_ + 
	C->rec_vec[1][j] * RP->I[i]->k_ +
	C->rec_vec[2][j] * RP->I[i]->l_ ;
    }

    k2 = dot(k,k);

    sqk2 = sqrt(k2);

    /* 
       6/98 NOTE: how can the argument be greater than 1 and make the
       arcsin NaN?  if the plane separation is smaller than the
       wavelength of light then it makes no sense to think of a
       scattered intensity.  this workaround makes the corresponding
       form factors zero, and hence the spot is effectively out of the
       calculation.  this should not be silent. we will put a warning
       message.  
    */
    if(RP->wavelength * sqk2 > 2.0)  {
      RP->xray_corr[i][j] = 0.0;
      push_generic_warning("get_correction", 
			   "skipping a spot with the interplanar spacing\n"
			   "\tsmaller than the wavelength");
      continue;
    }

    theta = asin(0.5 * RP->wavelength * sqk2);

    /* Polarization and Lorentz correction */

    if(RP->type == XRAY || RP->type == BOTH) {
      double m = cos(2.0 * theta);
      pol = 0.5 * (1.0 + m * m) / (sin(theta) * sin(2.0 * theta));
    } 

    /* if we have two types of radiation we have two angles */
    if(RP->type == BOTH) {
      theta = asin(0.5 * RP->wavelength2 * sqk2);
    }

    /* Lorentz factor for neutrons */
    if(RP->type == NEUTRON || RP->type == BOTH) {
      npol = 1.0 / (sin(theta) * sin(2.0 * theta));
    }

    k2 *= 0.25;

    aa = T->unique->head;
    for(j = 0; j < T->unique->length; j++) {
      unsigned m = aa->atom->SP_tag; /* we use this to access the scattering info */
      double Debye_Waller = aa->atom->DW_fac;
      double occ = aa->atom->occupancy;
      double * pp = SP[m]->params;   
 
      if(RP->type == XRAY || RP->type == BOTH) {
	RP->xray_corr[i][j] =
	  sqrt((double)RP->I[i]->multiplicity * pol) * 
	  /* here we need the occupancy and the Debye Waller factor  */
	  occ * exp( - Debye_Waller * k2 ) * 
	  ( pp[0] * exp( - pp[1] * k2 ) +
	    pp[2] * exp( - pp[3] * k2 ) +
	    pp[4] * exp( - pp[5] * k2 ) +
	    pp[6] * exp( - pp[7] * k2 ) +
	    pp[8] ); 
      } 
      if(RP->type == NEUTRON || RP->type == BOTH) {
	RP->neu_corr[i][j] =
	  sqrt((double) RP->I[i]->multiplicity * npol) * 
	  /* here we need the Debye Waller factor */
	  occ * exp(- Debye_Waller * k2) * pp[9];  
      }

      aa = aa->next;

    }
    
/* MWD Change */  
      for(o = 0, d = 0; o < T->Nmol; o++) {
	/* we need to get the scattering info for the
	   Molecule atoms as well 10/99 Marco Falcioni */
	int k;
	
	for(k = 0; k < T->M[o]->N; d++, k++) {
	  unsigned m = T->M[o]->A[k]->SP_tag; /* we use this to access
						 the scattering info */
/* MWD Change */
	  double Debye_Waller = T->M[o]->A[k]->DW_fac;
	  double occ = T->M[o]->A[k]->occupancy;
	  double * pp = SP[m]->params;   
	  
	  if(RP->type == XRAY || RP->type == BOTH) {
	    RP->xray_corr[i][d + T->n_unique] =
	      sqrt((double)RP->I[i]->multiplicity * pol) * 
	      /* here we need the occupancy and the Debye Waller factor  */
	      occ * exp( - Debye_Waller * k2 ) * 
	      ( pp[0] * exp( - pp[1] * k2 ) +
		pp[2] * exp( - pp[3] * k2 ) +
		pp[4] * exp( - pp[5] * k2 ) +
		pp[6] * exp( - pp[7] * k2 ) +
		pp[8] ); 
          }
	  
          if(RP->type == NEUTRON || RP->type == BOTH) {
            RP->neu_corr[i][d + T->n_unique] =
            sqrt((double) RP->I[i]->multiplicity * npol) * 
	     /* here we need the Debye Waller factor */
	    occ * exp(- Debye_Waller * k2) * pp[9];  
	  } 

	}
      }

  }



  return;
}
