/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html


*/
/* 

   implementation of interface specified in group.h

*/

#include <math.h>
#include <stdlib.h>
#include "constants.h"
#include "memory.h"
#include "error.h"
#include "matrices.h"
#include "group.h"

struct Group_el * new_Group_el (void)
{
  unsigned i,j;
  struct Group_el * This;

  This = (struct Group_el *) Malloc(sizeof(struct Group_el));

  if(This == 0) {
    push_generic_error("new_Group_el", "alloc failed"); 
  }                             
  This->W = new_matrix_d (3);
  This->w = new_vector_d (3);
  
  for(i = 0; i < 3; ++i) {
    for(j = 0; j < 3; ++j) {
      This->W[i][j] = 0.0;
    }
    This->w[i] = 0.0;
  }

  return This;

}

void delete_Group_el (struct Group_el * This)
{
  free_matrix_d( This->W, 3 );
  free_vector( This->w );
  Free(This);
  return;
}


/* 
   general rule for group multiplication 
   Ga = (A,a)
   Gb = (B,b)
   Gc = (C,c)

   y = Cx + c

   z = BCx + Bc + b 

   Ga = (BC, Bc + b) or A = BC, a = Bc + b


   multiplies two group elements and returns the first argument 
   ASSUMING it has been allocated already!
*/

void mult_Group_el (struct Group_el * a,
			const struct Group_el * b,
			const struct Group_el * c)
{
  unsigned long i;

  if(b == 0 || c == 0 || a == 0) {
    push_generic_error("multiply_Group_el","(nil) arg passed"); 
  }

  matrix_multiply_d(a->W, b->W, c->W, 3);
  matvec_multiply_d(a->w, b->W, c->w, 3);
  
  for(i = 0; i < 3; i++)
    a->w[i] += b->w[i];
  /* the following makes sure the translation vector is in the positive 
     direction and in the first cell. */
  for(i = 0; i < 3; i++)
    a->w[i] = fmod(a->w[i],1.0);
  
  return;
}

/* compare two group elements, and return the difference */

double comp_Group_el (const struct Group_el * a, const struct Group_el * b)
{
  double sum = 0.0;
  long i,j;
  double verytiny = 1.0001;

#ifdef DEBUG 
  write_Group_el(a, stdout);
  write_Group_el(b, stdout);
#endif
  if(a != 0 && b != 0){
    for(i = 0; i < 3; i++) {
      double t;
      for(j = 0; j < 3; j++) {
	sum += fabs(a->W[i][j] - b->W[i][j]);
      }
      t = fmod(a->w[i] -  b->w[i] + verytiny, 1.0);
      sum += fabs(t);
    }
    return sum;
  } else {
    push_generic_error("comp_Group_el","(nil) args passed");
    return max_double;
  }
}
/* 
   apply a group element to a point: the result is stored in a previously 
   allocated atom

   x' = Wx + w

*/
   
void apply_Group_el (const struct Group_el * g, 
		     const struct Atom * A, struct Atom * B)
{
  double ra,rb,rc, r[3];
  if(B == 0 || A == 0 || g == 0) {
    push_generic_error("apply_Group_el","(nil) arg passed");
  }
  
  ra = A->ra;
  rb = A->rb;
  rc = A->rc;

  r[0] = g->W[0][0] * ra + g->W[0][1] * rb + g->W[0][2] * rc + g->w[0];
  r[1] = g->W[1][0] * ra + g->W[1][1] * rb + g->W[1][2] * rc + g->w[1];
  r[2] = g->W[2][0] * ra + g->W[2][1] * rb + g->W[2][2] * rc + g->w[2];

  move_Atom(B,r);

  return;
}

/* 
   apply a group element to a point without translation
   the result is stored in a previously allocated atom

   x' = Wx

*/

void apply_Group_el_noT (const struct Group_el * g, 
			 const struct Atom * A, struct Atom * B)
{

  long i,j;
  double vec[3], r[3];
  if(B == 0 || A == 0 || g == 0) {
    push_generic_error("apply_Group_el","(nil) arg passed");
  }

  vec[0] = A->ra;
  vec[1] = A->rb;
  vec[2] = A->rc;

  for(i = 0; i < 3; i++) {
    double temp = g->W[i][0] * vec[0];
    for(j = 1; j < 3; j++){
      temp += g->W[i][j] * vec[j];
    }
    r[i] = temp;
  }

  move_Atom(B,r);

  return;
}

/* 
   this routine reads from stream a single Group element,
   written in the standard way of Deem and Newsam, as 12 floats in a line
   which is quite compact. It also allocates the group element
   comments are allowed after the 12 numbers
   White or commented lines preceding the 12 numbers are ignored graciously.
*/
struct Group_el * read_Group_el (FILE * stream)
{
  char line[1024], *p;
  struct Group_el * g = new_Group_el();

  p = get_line(line, sizeof(line), stream);
  
  if(p == 0 || sscanf(p,
	    "%lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
	    &(g->W[0][0]),&(g->W[0][1]),&(g->W[0][2]),
	    &(g->W[1][0]),&(g->W[1][1]),&(g->W[1][2]),
	    &(g->W[2][0]),&(g->W[2][1]),&(g->W[2][2]),
	    &(g->w[0]),   &(g->w[1]),  &(g->w[2])) != 12) {
    push_generic_error("read_Group_el",
		       "invalid data for symmetry element");
    exit_clean();
  } 

  return g;
}

/* read a line of a Symmetry specification in a XTL file */

struct Group_el * read_Group_el_XTL (FILE * stream)
{
  char line[1024], *p;
  char sym[5], mat[5];
  struct Group_el * g = new_Group_el();

  p = get_line(line, sizeof(line), stream);
  
  if(p == 0 || sscanf(p,
	    "%s %s %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf",
	    sym, mat,	      
	    &(g->W[0][0]),&(g->W[0][1]),&(g->W[0][2]),
	    &(g->W[1][0]),&(g->W[1][1]),&(g->W[1][2]),
	    &(g->W[2][0]),&(g->W[2][1]),&(g->W[2][2]),
	    &(g->w[0]),   &(g->w[1]),  &(g->w[2])) != 14) {
    push_generic_error("read_Group_el",
		       "invalid data for symmetry element");
    exit_clean();
  } 

  return g;
}

/* 
   writes to stream a single group element as 12 floats in a row
*/
void write_Group_el (const struct Group_el * g, FILE * stream)
{
  int i,j;
  if(stream == 0 || g == 0) {
    push_generic_error("write_Group_el", "(nil) args passed");
  }

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      fprintf(stream,"%4.1f ", g->W[i][j]);
  for(i=0;i<3;i++)
    fprintf(stream,"%8.6f ", g->w[i]);
  fprintf(stream,"\n");

}

void write_Group_el_XTL(const struct Group_el * g, FILE * stream)
{
  int i,j;
  if(stream == 0 || g == 0) {
    push_generic_error("write_Group_el_XTL", "(nil) args passed");
  }
  
  fprintf(stream, "SYM MAT");

  for(i=0; i<3; i++)
    for(j=0; j<3; j++)
      fprintf(stream," %4.1f", g->W[i][j]);
  for(i=0;i<3;i++)
    fprintf(stream," %8.6f", g->w[i]);
  fprintf(stream,"\n");
 
}

void copy_Group_el(struct Group_el * a, const struct Group_el * b)
{
  unsigned long i,j;
  if(a != 0 && b != 0) {
    for(i = 0; i < 3; i++) {
      for(j = 0; j < 3; j++)
	a->W[i][j] = b->W[i][j];
      
      a->w[i] = b->w[i];
    }
  } else {
    push_generic_error("copy_Group_el","(nil) args passed");
    exit_clean();
  }
  return;
}

