/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html


   Smart (faster) routine to compute the energy:
   we measure it exploiting symmetry: we only measure the energy of 
   the unique atoms and multiply that by the number of non merged atom in 
   each case.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "error.h"
#include "constants.h"
#include "interpolation.h"

#include "functions.h"
#include "misc.h"
#include "grid.h"
#include "chain.h"
#include "link.h"
#include "atoms.h"
#include "cell.h"
#include "positions.h"
#include "box.h"
#include "spline.h"
#include "params.h"
#include "verbose.h"
#include "conf.h"

double build_framework(const struct Params * P,
		       struct Chain * N,
		       struct Atom * A, double * ene, int * reject);

#undef DEBUG
#undef DEBUG2
#undef MWW

double energy(const struct Params * P, const struct Cell * C, 
	      const struct Positions * T, const struct Grid * Gr,
	      struct Conf * Co)
{

  unsigned long i, u_atom = 0;
  int reject = 0;
  struct Link * L;
  struct Chain * NNeigh;
  double tot_ene, en_merge;
  unsigned long * count;
  double * energy, * merge;
  double emem[6];  /* used to get detailed energy info to 
		      build_framework */

#ifndef NOCHECKS
  if(Gr == 0 || C == 0 || Gr->filled != 1) {
    push_generic_error("energy","(nil) or invalid args");
    exit_clean();
  }
#endif

  /* create neighborhood */
  NNeigh = new_Chain();

  count = (unsigned long *) malloc(T->n_unique * sizeof(unsigned long));
  energy = (double *) malloc(T->n_unique * sizeof(double));
  merge = (double *) malloc(T->n_unique * sizeof(double));

  for(i = 0; i < T->n_unique; i++) {
    count[i] = 0;
    merge[i] = 0.0;
    energy[i] = 0.0;
  }

  /* 
     here we compute the contribution from the expected density
     basically we want the actual number of T atoms to be a given amount 
  */

  en_merge = ((double)T->density - (double) P->n_actual);
  en_merge = en_merge * en_merge;

  /* we fill the neighborhoods */
  for(L = T->unique->head; L != 0; L = L->next) {
    struct Box * Box;
    struct Atom * A = L->atom;
    struct Link * N;

    Box = index_atom_Grid(Gr, A, C, 0); /* box where A lives */
    /* this info may eventually be cached in the Atom to save this call */

    /* first loop through atoms in the same box */
    for(N = Box->list->head; N != 0; N = N->next) {
      double dist2;
      if(A == N->atom) continue;
      dist2 = dist2_xyz(C, A, N->atom);
      if(dist2 < P->frange2) {
	struct Link * newlink = new_Link(N->atom);
	xyz_Atom(N->atom,C);
	newlink->key = sqrt(dist2);
	add_Link(newlink, NNeigh);
      }
    }

    /* loop through atoms in the neighboring boxes */
    for(i = 0; i < Box->n; i++) {
      for(N = Box->near[i]->list->head; N != 0; N = N->next) {
	double dist2 = dist2_xyz(C, A, N->atom);
	if(dist2 < P->frange2) {
	  struct Link * newlink = new_Link(N->atom);
	  xyz_Atom(N->atom,C);
	  newlink->key = sqrt(dist2);
	  add_Link(newlink, NNeigh);
	}
      }
    }

    /* take into account merging energy */

    if(A->merge_d > 0.0) {
      merge[u_atom] = A->merge_n * 
	(P->ene_merge * ( 2.0 * sqrt(A->merge_d / P->merge2) - 1.0));
#ifdef DEBUG
      printf("merging energy %f distance %f %f\n",merge[u_atom],A->merge_d,
	     P->merge2);
      fflush(stdout);
#endif
      Co->ene[6][u_atom] = merge[u_atom];
    } else {
      Co->ene[6][u_atom] = 0.0;
    }

    if(NNeigh->length == 0) {
      /* this is a lonely atom indeed! */
      energy[u_atom] = P->wt_unbond * 1000.0;
 /* MWD Change */
      {int jj;
        for (jj=0; jj<6; jj++) emem[jj] = 0.0;
      } 
    } else {
      energy[u_atom] = build_framework(P, NNeigh, A, emem, &reject);
    }

    if(reject) {
      tot_ene =  DBL_MAX;
      Co->Pos_energy = DBL_MAX;
      goto end;
    }

#ifdef DEBUG
    printf("energy_framework = %f atom %d\n",energy[u_atom], u_atom);
#endif

    Co->ene[0][u_atom] = emem[0];
    Co->ene[1][u_atom] = emem[1];
    Co->ene[2][u_atom] = emem[2];
    Co->ene[3][u_atom] = emem[3];
    Co->ene[4][u_atom] = emem[4];
    Co->ene[5][u_atom] = emem[5];

    Co->pos[0][u_atom] = A->ra;
    Co->pos[1][u_atom] = A->rb;
    Co->pos[2][u_atom] = A->rc;

    u_atom++;
    clean_Chain(NNeigh);
  }

  for(L = T->distinct->head; L != 0; L = L->next) {
    count[L->atom->root]++;
  }

  /* merging term */
  tot_ene = P->wt_merge * en_merge;


#ifdef DEBUG2
  printf("EN %e", tot_ene);
#endif
  for(i = 0; i < T->n_unique; i++) {
    double scra = (double) count[i]/(double) P->n_actual;
    double scra2 = (double) (T->n_symmetry - count[i])/ (double) P->n_actual;
    Co->ene[7][i] = (double) count[i];

    /* PATCH FOR MWW GREAT UGLINESS */ 
#ifdef MWW
    if(count[i] < 4) {
      tot_ene =  DBL_MAX;
      Co->Pos_energy = DBL_MAX;
      goto end;
    } 
#endif
    tot_ene += scra * energy[i] + scra2 * merge[i];

#ifdef DEBUG2
  printf(" %e %e", count[i] * energy[i], (T->n_symmetry - count[i])* merge[i]);
#endif   
  }
#ifdef DEBUG2
  printf("\n");
  fflush(stdout);
#endif  

  Co->Pos_energy = tot_ene;

 end: 
  clean_Chain(NNeigh); /* just in case */
  delete_Chain(NNeigh);
  free(count);
  free(energy);
  free(merge);

  return tot_ene;
}

